# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

# Import locale from sclapp before all other imports to work around bug in
# stdlib.
# See http://www.selenic.com/mercurial/wiki/index.cgi/Character_Encoding_On_OSX.
from sclapp import locale

import sys, os, errno, fcntl, re, shutil, platform

try:
    from functools import wraps
except ImportError:
    from sclapp.legacy_support import wraps

from optparse import OptParseError, SUPPRESS_HELP
from threading import Lock, RLock

import fuse
from fuse import Fuse

from pytagsfs.optgroup import GroupingOptionParser
from pytagsfs.subspat import SubstitutionPattern
from pytagsfs.sourcetree import SourceTree

from pytagsfs.metastore import DelegateMultiMetaStore

from pytagsfs.sourcetreemon import (
  SOURCE_TREE_MONITORS,
  get_source_tree_monitor,
)
from pytagsfs.exceptions import (
  PathNotFound,
  InvalidArgument,
  MissingDependency,
)
from pytagsfs.pathpropcache import PathPropCache
from pytagsfs.debug import (
  logged,
  verbosely_logged,
  exceptionally_logged,
  log_debug,
  log_warning,
  log_loud,
  set_log_level,
  DEBUG,
  INFO,
  WARNING,
  ERROR,
  CRITICAL,
  enable_stderr,
  set_logsize,
)
from pytagsfs.util import (
  get_obj_by_dotted_name,
  nonatomic,
  attr_named,
  return_errno,
  stat_result_to_fuse_stat,
  split_path,
  unicode_path_sep,
)

from pytagsfs.specialfile import SpecialFileFuseMixin
from pytagsfs.specialfile.logfile import VirtualLogFile
from pytagsfs.profiling import (
  profiled_fs_method,
  profiled_file_method,
  profiled_file_method_noargs,
  enable_profiling,
)

from pytagsfs import __version__ as version

fuse.fuse_python_api = (0, 2)

DEFAULT_VERBOSITY = 'warning'


# I'll try to make a long story short:

# With FUSE filesystems, direct_io is incompatible with mmap as the size of the
# file is not always known (it may change arbitrarily).  On Linux, mmap calls
# immediately return ENODEV, and the application has an opportunity to fall
# back to some other reasonable behavior.  However, on Darwin, the initial mmap
# call succeeds, and the following attempts to modify the mmap'ed file end in
# SIGBUS (from what I've seen), which is non-catchable.  Thus, the application
# immediately terminates, and, unless the application has been coded very
# carefully, the file may be corrupted.  Since users can't be expected to know
# if their application uses mmap or not, we must disable write support.

# Our approach is to track platforms that are known-good and platforms that are
# known-bad.  All others will issue a warning that write support is experimental
# and they may encounter file corruption.

# Platforms with known-bad mmap behavior:
MMAP_KNOWN_BAD_PLATFORMS = ('Darwin',)

# Platforms with known-good mmap behavior:
MMAP_KNOWN_GOOD_PLATFORMS = ('Linux',)


PLATFORM = platform.system()


if PLATFORM in ('FreeBSD', 'NetBSD', 'OpenBSD', 'Darwin'):
    READ_ONLY_MOUNT_OPTION = 'rdonly'
else:
    READ_ONLY_MOUNT_OPTION = 'ro'


def append_filter(option, opt_str, value, parser, *args, **kwargs):
    if not hasattr(parser.values, 'filters'):
        parser.values.filters = []
    if 'src' in opt_str:
        which = 'src'
    else:
        which = 'dst'
    parser.values.filters.append((which, value))


def operation_on_one_real_path(real_op):
    @return_errno
    @wraps(real_op)
    def wrapper(self, fake_path, *args, **kwargs):
        real_path = self.encode_real_path(self.source_tree_rep.get_real_path(
          self.decode_fake_path(fake_path)))
        return real_op(real_path, *args, **kwargs)
    return wrapper


def operation_on_two_real_paths(real_op):
    @return_errno
    @wraps(real_op)
    def wrapper(self, fake_path1, fake_path2, *args, **kwargs):
        real_path1 = self.encode_real_path(self.source_tree_rep.get_real_path(
          self.decode_fake_path(fake_path1)))
        real_path2 = self.encode_real_path(self.source_tree_rep.get_real_path(
          self.decode_fake_path(fake_path2)))
        return real_op(real_path1, real_path2, *args, **kwargs)
    return wrapper


class FileSystemMappingToRealFilesOptionParser(GroupingOptionParser):
    DEFAULT_OPTIONS = {}
    DEFAULT_OPTIONS['-d'] = {
      'default': False,
      'action': 'store_true',
      'help': 'debug mode; implies -f',
    }
    DEFAULT_OPTIONS['-f'] = {
      'default': False,
      'action': 'store_true',
      'help': 'stay in foreground',
    }
    DEFAULT_OPTIONS['-s'] = {
      'default': False,
      'action': 'store_true',
      'help': 'disable multi-threaded operation',
    }
    DEFAULT_OPTIONS['-r'] = {
      'default': False,
      'action': 'store_true',
      'help': 'mount read-only',
    }

    DEFAULT_OPTION_ORDER = ('-f', '-d', '-s')

    DEFAULT_MOUNT_OPTIONS = {}
    DEFAULT_MOUNT_OPTIONS[READ_ONLY_MOUNT_OPTION] = dict(DEFAULT_OPTIONS['-r'])
    DEFAULT_MOUNT_OPTIONS['force_write_support'] = {
      'default': False,
      'action': 'store_true',
      'help': 'force write support, even on broken platforms',
    }
    DEFAULT_MOUNT_OPTIONS['iocharset'] = {
      'default': 'utf-8',
      'metavar': 'ENCODING',
      'help':
        'set mounted tree character encoding to ENCODING (default: %default)',
    }
    DEFAULT_MOUNT_OPTIONS['source_iocharset'] = {
      'default': 'utf-8',
      'metavar': 'ENCODING',
      'help':
        'set source directory character encoding to ENCODING '
        '(default: %default)',
    }
    DEFAULT_MOUNT_OPTIONS['srcfilter'] = {
      'action': 'callback',
      'type': 'str',
      'callback': append_filter,
      'metavar': 'EXPR',
      'help': 'add a source path filter as specified by EXPR',
    }
    DEFAULT_MOUNT_OPTIONS['dstfilter'] = {
      'action': 'callback',
      'type': 'str',
      'callback': append_filter,
      'metavar': 'EXPR',
      'help': 'add a destination path filter as specified by EXPR',
    }
    DEFAULT_MOUNT_OPTIONS['debug'] = {
      'default': False,
      'action': 'store_true',
      'help': 'synonym for -d',
    }
    DEFAULT_MOUNT_OPTIONS['verbosity'] = {
      'default': 'warning',
      'help':
        'log file verbosity; one of debug, info, warning, error, critical '
        '(default: %default)',
    }
    DEFAULT_MOUNT_OPTIONS['logsize'] = {
      'type': 'int',
      'default': (1024 * 1024),
      'help': 'length of log file ring buffer, in bytes',
    }
    DEFAULT_MOUNT_OPTIONS['allow_other'] = {
      'action': 'store_true',
      'default': False,
      'help': 'allow access to other users',
    }
    DEFAULT_MOUNT_OPTIONS['allow_root'] = {
      'action': 'store_true',
      'default': False,
      'help': 'allow access to root',
    }
    DEFAULT_MOUNT_OPTIONS['nonempty'] = {
      'action': 'store_true',
      'default': False,
      'help': 'allow mounts over non-empty file/dir',
    }
    DEFAULT_MOUNT_OPTIONS['uid'] = {'help': 'set owner to UID (if supported)'}
    DEFAULT_MOUNT_OPTIONS['gid'] = {'help': 'set group to GID (if supported)'}
    DEFAULT_MOUNT_OPTIONS['profile'] = {
      'action': 'store_true',
      'default': False,
      'help': 'log profiling data',
    }

    DEFAULT_MOUNT_OPTION_ORDER = (
      'srcfilter',
      'dstfilter',
      'iocharset',
      'source_iocharset',
      READ_ONLY_MOUNT_OPTION,
      'allow_other',
      'allow_root',
      'nonempty',
      'uid',
      'gid',
      'verbosity',
      'logsize',
      'debug',
      'profile',
    )

    def __init__(self, *args, **kwargs):
        kwargs['usage'] = '%prog [OPTIONS] {source} {mountpoint}'
        kwargs['version'] = '%%prog version %s' % version
        GroupingOptionParser.__init__(self, *args, **kwargs)
        self.add_default_options()

    def add_default_options(self):
        for opt in self.DEFAULT_OPTION_ORDER:
            kwargs = self.DEFAULT_OPTIONS[opt]
            self.add_option(opt, **kwargs)

        for opt, kwargs in self.DEFAULT_OPTIONS.items():
            if opt not in self.DEFAULT_OPTION_ORDER:
                self.add_option(opt, **kwargs)

        self.add_group(
          '-o',
          title = 'Mount Options',
          help = "mount options (see `Mount Options')",
        )

        for opt in self.DEFAULT_MOUNT_OPTION_ORDER:
            kwargs = self.DEFAULT_MOUNT_OPTIONS[opt]
            self.add_mount_option(opt, **kwargs)

        for opt, kwargs in self.DEFAULT_MOUNT_OPTIONS.items():
            if opt not in self.DEFAULT_MOUNT_OPTION_ORDER:
                self.add_mount_option(opt, **kwargs)

    def add_mount_option(self, option, **kwargs):
        if 'group' in kwargs:
            raise AssertionError
        return self.add_option(option, group = '-o', **kwargs)

    def parse_args(self, *args, **kwargs):
        retval = GroupingOptionParser.parse_args(self, *args, **kwargs)
        if not hasattr(self.values.o, 'filters'):
            self.values.o.filters = []
        return retval


fuse.feature_assert('stateful_files')

class File(object):
    _lock = None

    keep_cache = False

    # Undefined File methods:
    #
    # * lock

    def __init__(
      self, filesystem_instance, fake_path, flags, mode = None,
      truncate_to = None
    ):
        self._lock = RLock()
        self.filesystem_instance = filesystem_instance
        self.fake_path = fake_path
        self.real_path = filesystem_instance.source_tree_rep.get_real_path(
          fake_path)
        self.flags = flags
        self.mode = mode
        self.truncate_to = truncate_to

    def fgetattr(self): pass
    def read(self, length, offset): pass
    def release(self, flags): pass
    def write(self, buf, offset): pass
    def fsync(self, isfsyncfile): pass
    def flush(self): pass
    def ftruncate(self, len): pass


class ReadOnlyFile(File):
    fd = None
    file = None

    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, *args, **kwargs):
        super(ReadOnlyFile, self).__init__(*args, **kwargs)
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        args = [real_path, self.flags]
        if self.mode is not None:
            args.append(self.mode)
        self.file = os.fdopen(os.open(*args), 'r')

    @profiled_file_method
    @exceptionally_logged
    @logged
    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        self.file.seek(offset)
        return self.file.read(length)

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        return stat_result_to_fuse_stat(os.fstat(self.file.fileno()))

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        self.file.close()

    @profiled_file_method_noargs
    @exceptionally_logged
    @logged
    def write(self, buf, offset):
        raise InvalidArgument

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    def ftruncate(self, len):
        raise InvalidArgument


class ReadWriteFile(File):
    fd = None

    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, *args, **kwargs):
        super(ReadWriteFile, self).__init__(*args, **kwargs)
        self.copy_path = os.path.join(
          os.path.dirname(self.real_path),
          u'.pytagsfs.%s' % os.path.basename(self.real_path)
        )
        self.open_file()

    @nonatomic(attr_named('_lock'))
    def open_file(self):
        flags = self.flags
        flags = flags | os.O_CREAT | os.O_EXCL

        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        copy_path = self.filesystem_instance.encode_real_path(self.copy_path)

        args = [copy_path, flags]
        if self.mode is not None:
            args.append(self.mode)

        log_debug(
          u'ReadWriteFile.open_file: calling os.open with args %s' % repr(args))
        self.fd = os.open(*args)

        f = open(real_path, 'r')
        if self.truncate_to is None:
            data = f.read()
        else:
            data = f.read(self.truncate_to)
        f.close()

        os.write(self.fd, data)
        os.lseek(self.fd, 0, 0)

        shutil.copystat(real_path, copy_path)

    @profiled_file_method
    @exceptionally_logged
    @logged
    @nonatomic(attr_named('_lock'))
    def read(self, length, offset):
        os.lseek(self.fd, offset, 0)
        return os.read(self.fd, length)

    @profiled_file_method_noargs
    @exceptionally_logged
    @logged
    @nonatomic(attr_named('_lock'))
    def write(self, buf, offset):
        os.lseek(self.fd, offset, 0)
        return os.write(self.fd, buf)

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def release(self, flags):
        retval = os.close(self.fd)
        real_path = self.filesystem_instance.encode_real_path(self.real_path)
        copy_path = self.filesystem_instance.encode_real_path(self.copy_path)
        os.rename(copy_path, real_path)
        return retval

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fsync(self, isfsyncfile):
        if isfsyncfile and hasattr(os, 'fdatasync'):
            return os.fdatasync(self.fd)
        else:
            return os.fsync(self.fd)

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def flush(self):
        return os.close(os.dup(self.fd))

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def fgetattr(self):
        return os.fstat(self.fd)

    @profiled_file_method
    @exceptionally_logged
    @verbosely_logged
    @nonatomic(attr_named('_lock'))
    def ftruncate(self, len):
        return os.ftruncate(self.fd, len)


class ModeSelectFile(File):
    @exceptionally_logged(ignore = (PathNotFound,))
    @verbosely_logged
    def __init__(self, filesystem_instance, fake_path, flags, mode = None,
      truncate_to = None):
        if flags & (os.O_WRONLY | os.O_RDWR):
            cls = ReadWriteFile
        else:
            cls = ReadOnlyFile

        self.__class__ = cls
        cls.__init__(self, filesystem_instance, fake_path, flags, mode = mode,
          truncate_to = truncate_to)


class FileSystemMappingToRealFiles(Fuse):
    _lock = None

    user_encoding = None

    source_tree_rep = None
    source_tree_path = None
    iocharset = None
    source_iocharset = None
    path_filters = None
    debug = False
    foreground = False
    readonly = None
    force_write_support = None
    verbosity = None
    logsize = None

    truncated_paths = None
    write_locks = None
    write_locks_lock = None

    file_instances = None

    file_class = ModeSelectFile

    subtype = None

    verbosities = {
      'debug': DEBUG,
      'info': INFO,
      'warning': WARNING,
      'error': ERROR,
      'critical': CRITICAL,
    }

    cmdline_arguments = None
    cmdline_parser = None
    fuse_cmdline_arguments = None

    @exceptionally_logged
    @logged
    def __init__(self, *args, **kwargs):
        self._lock = RLock()

        kwargs['dash_s_do'] = 'setsingle'

        self.truncated_paths = {}
        self.write_locks = {}

        self.file_instances = {}

        if self.cmdline_parser is None:
            self.cmdline_parser = FileSystemMappingToRealFilesOptionParser()

        super(FileSystemMappingToRealFiles, self).__init__(*args, **kwargs)

    fuse.feature_assert('has_init')
    @exceptionally_logged
    @logged
    def fsinit(self):
        # We prefer to call source_tree_rep.start prior to the filesystem
        # process detaching from the user's shell, since the initial scan of
        # the source files can take some time and detaching before it is
        # complete leads to confusion as to whether the filesystem is hung or
        # not.  Some SourceTreeMonitor implementations can't carry across the
        # fork call, though, so this provides a mechanism for start to be called
        # after that transition occurs.  If this is not the case,
        # source_tree_rep.start is called at the end of pre_init, rather than
        # here.

        if not self.source_tree_rep.can_handle_fork():
            log_warning(
              u"source tree representation can't handle fork, "
              u'calling source_tree_rep.start in fsinit'
            )
            self.source_tree_rep.start()

    # This clean-up is performed at the end of main to work around what appears
    # to be a bug in python-fuse that prevents exit due to a GIL deadlock.  The
    # bug is fixed upstream, but the fix may not have been released yet.

    #fuse.feature_assert('has_destroy')
    #@exceptionally_logged
    #def fsdestroy(self):
    #    self.source_tree_rep.stop()

    @exceptionally_logged
    @logged
    def main(self):
        # locale.getpreferredencoding is not necessarily thread-safe, so we
        # call it here before any threads might be forked:
        self.user_encoding = locale.getpreferredencoding()

        try:
            self.parse()
        except OptParseError, e:
            self.cmdline_parser.error(unicode(e))

        try:
            retval = super(FileSystemMappingToRealFiles, self).main()
        except fuse.FuseError, e:
            log_loud(unicode(e))
            retval = 1
        else:
            # See note regarding fsdestroy.
            self.source_tree_rep.stop()

        return retval

    def parse(self):
        self.cmdline_arguments = sys.argv[1:]
        opts, args = self.cmdline_parser.parse_args(self.cmdline_arguments)

        if len(args) < 2:
            raise OptParseError('too few arguments')
        if len(args) > 2:
            raise OptParseError('too many arguments')

        self.fuse_cmdline_arguments = [
          '-o', 'default_permissions',
          '-o', 'entry_timeout=0',
          '-o', 'negative_timeout=0',
          '-o', 'attr_timeout=0',
          '-o', 'direct_io',
        ]

        if PLATFORM == 'Darwin':
            self.fuse_cmdline_arguments.extend([
              '-o', 'nolocalcaches',
              '-o', 'noreadahead',
              '-o', 'noubc',
              '-o', 'novncache',
            ])

        if self.subtype:
            self.fuse_cmdline_arguments.extend(
              ['-o', 'subtype=%s' % self.subtype])

        if opts.r or getattr(opts.o, READ_ONLY_MOUNT_OPTION):
            self.fuse_cmdline_arguments.extend(['-o', READ_ONLY_MOUNT_OPTION])

        if opts.d or opts.o.debug:
            self.fuse_cmdline_arguments.append('-d')

        if opts.f:
            self.fuse_cmdline_arguments.append('-f')

        if opts.s:
            self.fuse_cmdline_arguments.append('-s')

        if opts.o.allow_other:
            self.fuse_cmdline_arguments.extend(['-o', 'allow_other'])

        if opts.o.allow_root:
            self.fuse_cmdline_arguments.extend(['-o', 'allow_root'])

        if opts.o.nonempty:
            self.fuse_cmdline_arguments.extend(['-o', 'nonempty'])

        if opts.o.uid is not None:
            self.fuse_cmdline_arguments.extend(
              ['-o', 'uid=%s' % str(opts.o.uid)])

        if opts.o.gid is not None:
            self.fuse_cmdline_arguments.extend(
              ['-o', 'gid=%s' % str(opts.o.gid)])

        self.source_tree_path = os.path.abspath(
          unicode(args[0], self.user_encoding))

        self.fuse_cmdline_arguments.extend([
          '-o', 'fsname=%s' % self.source_tree_path.encode(self.user_encoding)])

        self.fuse_cmdline_arguments.extend(args)

        retval = super(
          FileSystemMappingToRealFiles, self).parse(self.fuse_cmdline_arguments)

        self.process_options(opts, args)
        self.pre_init()
        return retval

    def process_options(self, opts, args):
        if opts.d or opts.o.debug:
            self.debug = True
            self.foreground = True
            self.verbosity = self.verbosities['debug']

        if opts.f:
            self.foreground = True

        if opts.r or getattr(opts.o, READ_ONLY_MOUNT_OPTION):
            self.readonly = True

        if opts.o.force_write_support:
            self.force_write_support = True

        if opts.o.verbosity is not None:
            if self.debug:
                log_warning(
                  u'warning: option debug overrides user-specified verbosity')
            else:
                try:
                    self.verbosity = self.verbosities[opts.o.verbosity]
                except KeyError:
                    raise OptParseError(
                      'verbosity should be one of %s' % ', '.join(
                        self.verbosities.keys())
                    )
        elif not self.debug:
            self.verbosity = self.verbosities[DEFAULT_VERBOSITY]

        self.logsize = opts.o.logsize

        self.iocharset = opts.o.iocharset
        self.source_iocharset = opts.o.source_iocharset

        if opts.o.profile:
            enable_profiling()

        self.path_filters = []
        for which, expr in opts.o.filters:
            real = (which == 'src')
            self.path_filters.append((expr, real))

    def pre_init(self):
        set_logsize(self.logsize)

        set_log_level(self.verbosity)
        if self.foreground:
            enable_stderr()

        log_warning(u'pytagsfs version %s' % version)
        log_warning(u'command-line arguments: %s' % str(self.cmdline_arguments))
        log_warning(
          u'fuse command-line arguments: %s' % unicode(
            self.fuse_cmdline_arguments
        ))

        self.source_tree_rep = self.build_source_tree_rep()
        self.multithreaded = (
          self.multithreaded and
          self.source_tree_rep.supports_threads()
        )
        log_warning(u'multithreaded = %s' % self.multithreaded)

        # See note in fsinit.
        if self.source_tree_rep.can_handle_fork():
            log_warning(
              u'source tree representation can handle fork, '
              u'calling source_tree_rep.start in pre_init'
            )
            self.source_tree_rep.start()

    def build_source_tree_rep(self):
        raise NotImplementedError

    @nonatomic(attr_named('_lock'))
    def write_lock_path(self, fake_path):
        lock = self.write_locks.get(fake_path, Lock())
        self.write_locks[fake_path] = lock
        lock.acquire()

    @nonatomic(attr_named('_lock'))
    def write_unlock_path(self, fake_path):
        self.write_locks[fake_path].release()

    def encode_fake_path(self, s):
        return s.encode(self.iocharset)

    def decode_fake_path(self, s):
        return s.decode(self.iocharset)

    def encode_real_path(self, s):
        return self.source_tree_rep.source_tree.encode(s)

    def decode_real_path(self, s):
        return self.source_tree_rep.source_tree.decode(s)

################################################################################

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def getattr(self, fake_path):
        fake_path = self.decode_fake_path(fake_path)
        try:
            return self.file_instances[fake_path].fgetattr()
        except KeyError:
            return stat_result_to_fuse_stat(
              self.source_tree_rep.getattr(fake_path))

    chmod = profiled_fs_method(exceptionally_logged(
      operation_on_one_real_path(os.chmod)
    ))
    chown = profiled_fs_method(exceptionally_logged(
      operation_on_one_real_path(os.chown)
    ))

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def truncate(self, fake_path, len):
        fake_path = self.decode_fake_path(fake_path)
        log_debug(u'truncate: write-locking path %s' % fake_path)
        self.write_lock_path(fake_path)
        try:
            try:
                if len < self.truncated_paths[fake_path]:
                    self.truncated_paths[fake_path] = len
            except KeyError:
                self.truncated_paths[fake_path] = len
        finally:
            log_debug(u'truncate: write-unlocking path %s' % fake_path)
            self.write_unlock_path(fake_path)

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def open(self, fake_path, flags):
        fake_path = self.decode_fake_path(fake_path)
        if (os.O_RDWR | os.O_WRONLY) & flags:
            log_debug(u'open: write-locking path %s' % fake_path)
            self.write_lock_path(fake_path)
        file_instance = ModeSelectFile(
          self, fake_path, flags,
          truncate_to = self.truncated_paths.get(fake_path, None)
        )
        if (os.O_RDWR | os.O_WRONLY) & flags:
            self.file_instances[fake_path] = file_instance
        return file_instance

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def release(self, fake_path, flags, file_obj):
        fake_path = self.decode_fake_path(fake_path)
        try:
            del self.truncated_paths[fake_path]
        except KeyError:
            pass
        retval = Fuse.__getattr__(self, 'release')(fake_path, flags, file_obj)
        if (os.O_RDWR | os.O_WRONLY) & flags:
            log_debug(u'release: write-unlocking path %s' % fake_path)
            self.write_unlock_path(fake_path)
        try:
            del self.file_instances[fake_path]
        except KeyError:
            pass
        return retval

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def create(self, fake_path, flags, mode):
        # FIXME: Is EINVAL correct?  Would the default (ENOSYS) be better?
        raise InvalidArgument

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def utime(self, fake_path, times):
        fake_path = self.decode_fake_path(fake_path)
        self.source_tree_rep.utime(fake_path, times)
        return 0

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    def statfs(self):
        return os.statvfs(self.source_tree_path)

    @profiled_fs_method
    @exceptionally_logged
    @logged
    @return_errno
    def readdir(self, fake_path, offset):
        fake_path = self.decode_fake_path(fake_path)
        entries = [
          self.encode_fake_path(e)
          for e in self.source_tree_rep.get_entries(fake_path)
        ]
        return [fuse.Direntry(e) for e in entries]

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def rename(self, old_fake_path, new_fake_path):
        old_fake_path = self.decode_fake_path(old_fake_path)
        new_fake_path = self.decode_fake_path(new_fake_path)
        self.source_tree_rep.rename_path(old_fake_path, new_fake_path)
        return 0

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def mkdir(self, fake_path, mode):
        fake_path = self.decode_fake_path(fake_path)
        self.source_tree_rep.add_directory(fake_path)
        return 0

    @profiled_fs_method
    @exceptionally_logged
    @verbosely_logged
    @return_errno
    @nonatomic(attr_named('_lock'))
    def rmdir(self, fake_path):
        fake_path = self.decode_fake_path(fake_path)
        self.source_tree_rep.remove_directory(fake_path)
        return 0

    # Undefined FUSE methods:
    #
    # * setxattr
    # * getxattr
    # * listxattr
    # * removexattr
    # * opendir
    # * releasedir
    # * fsyncdir
    # * init
    # * destroy
    # * utimens
    # * bmap
    # * readlink
    # * mknod
    # * unlink
    # * symlink
    # * link


################################################################################


class PyTagsFileSystemOptionParser(FileSystemMappingToRealFilesOptionParser):
    DEFAULT_MOUNT_OPTIONS = dict(
      FileSystemMappingToRealFilesOptionParser.DEFAULT_MOUNT_OPTIONS)

    DEFAULT_MOUNT_OPTIONS['sourcetreerep'] = {
      'default': 
        'pytagsfs.sourcetreerep.pollinline.PollInLineSourceTreeRepresentation',
      'help': SUPPRESS_HELP,
    }
    DEFAULT_MOUNT_OPTIONS['pathstore'] = {
      'default': 'pytagsfs.pathstore.pytypes.PyTypesPathStore',
      'help': SUPPRESS_HELP,
    }
    DEFAULT_MOUNT_OPTIONS['sourcetreemon'] = {
      'default': None,
      'help': SUPPRESS_HELP,
    }
    DEFAULT_MOUNT_OPTIONS['metastores'] = {
      'default': ';'.join([
        'pytagsfs.metastore.path.PathMetaStore',
        'pytagsfs.metastore.mutagen_.MutagenFileMetaStore',
      ]),
      'help': SUPPRESS_HELP,
    }
    DEFAULT_MOUNT_OPTIONS['format'] = {
      'default': u'%s%%f' % unicode_path_sep,
      'metavar': 'FORMAT',
      'help': 'set destination path format string (default: %default)',
    }
    DEFAULT_MOUNT_OPTIONS['nocache'] = {
      'action': 'store_true',
      'default': False,
      'help': 'disable path property caching',
    }

    DEFAULT_MOUNT_OPTION_ORDER = (
      'format',
      'srcfilter',
      'dstfilter',
      'iocharset',
      'source_iocharset',
      READ_ONLY_MOUNT_OPTION,
      'allow_other',
      'allow_root',
      'nonempty',
      'uid',
      'gid',
      'nocache',
      'verbosity',
      'logsize',
      'debug',
    )


class PyTagsFileSystem(
  SpecialFileFuseMixin, FileSystemMappingToRealFiles):
    special_file_classes = {
      '.log': VirtualLogFile,
    }

    source_tree_rep_cls_dotted_name = None
    path_store_cls_dotted_name = None
    source_tree_mon_cls_dotted_name = None
    meta_store_cls_dotted_names = None
    nocache = None

    format_string = None
    format_string_parts = None

    subtype = 'pytagsfs'

    def __init__(self, *args, **kwargs):
        if self.cmdline_parser is None:
            self.cmdline_parser = PyTagsFileSystemOptionParser()
        super(PyTagsFileSystem, self).__init__(*args, **kwargs)

    def process_options(self, opts, args):
        super(PyTagsFileSystem, self).process_options(opts, args)

        self.source_tree_rep_cls_dotted_name = opts.o.sourcetreerep
        self.path_store_cls_dotted_name = opts.o.pathstore
        self.source_tree_mon_cls_dotted_name = opts.o.sourcetreemon
        self.meta_store_cls_dotted_names = opts.o.metastores
        self.nocache = opts.o.nocache

        self.format_string = opts.o.format

        # format string must not end with a /
        self.format_string = self.format_string.rstrip(unicode_path_sep)

        # format_string must start with a /
        if not self.format_string.startswith(unicode_path_sep):
            self.format_string = '%s%s' % (
              unicode_path_sep, self.format_string)

        # format_string should not contain multiple consecutive /'s
        self.format_string = re.sub(
          ur'%s{2,}' % unicode_path_sep, unicode_path_sep, self.format_string)

        self.format_string_parts = split_path(self.format_string)

    @classmethod
    def get_meta_store(cls, meta_store_cls_dotted_names):
        meta_store_cls_dotted_names = meta_store_cls_dotted_names.split(';')
        meta_store_classes = [
          get_obj_by_dotted_name(cls_name)
          for cls_name in meta_store_cls_dotted_names
        ]
        if len(meta_store_classes) == 1:
            meta_store = meta_store_classes[0]()
        else:
            meta_store = DelegateMultiMetaStore([
              cls() for cls in meta_store_classes])
        return meta_store

    def build_source_tree_rep(self):
        substitution_patterns = [
          SubstitutionPattern(p) for p in self.format_string_parts
        ]

        source_tree_rep_cls = get_obj_by_dotted_name(
          self.source_tree_rep_cls_dotted_name)
        path_store_cls = get_obj_by_dotted_name(
          self.path_store_cls_dotted_name)

        if self.source_tree_mon_cls_dotted_name is not None:
            source_tree_mon_cls_dotted_names = [
              self.source_tree_mon_cls_dotted_name]
        else:
            source_tree_mon_cls_dotted_names = list(SOURCE_TREE_MONITORS)

        monitor = None
        for source_tree_mon_cls_dotted_name in (
          source_tree_mon_cls_dotted_names):
            log_debug(
              'trying source tree monitor class %s' %
              source_tree_mon_cls_dotted_name
            )
            try:
                monitor = get_source_tree_monitor(
                  source_tree_mon_cls_dotted_name)
            except MissingDependency, e:
                log_warning(
                  (
                    u'WARNING: not using source tree monitor %s due to '
                    u'missing dependency %s'
                  ) % (
                    source_tree_mon_cls_dotted_name,
                    unicode(e),
                  )
                )
            else:
                break

        if monitor is None:
            raise OptParseError('No supported source tree monitor found.')

        meta_store = self.get_meta_store(self.meta_store_cls_dotted_names)

        if self.nocache:
            cache = None
        else:
            cache = PathPropCache()

        log_warning(
          u'source_tree_rep_cls: %s.%s' % (
            source_tree_rep_cls.__module__, source_tree_rep_cls.__name__
        ))
        log_warning(
          u'path_store_cls: %s.%s' % (
            path_store_cls.__module__, path_store_cls.__name__
        ))
        log_warning(
          u'source_tree_mon_cls: %s.%s' % (
            monitor.__class__.__module__, monitor.__class__.__name__
        ))

        if not self.readonly:
            if not monitor.supports_writes():
                raise OptParseError(
                  ('Source tree monitor does not support writes. '
                   'You must mount read-only with -o %s.'
                  ) % READ_ONLY_MOUNT_OPTION
                )

            if self.force_write_support:
                if PLATFORM in MMAP_KNOWN_BAD_PLATFORMS:
                    log_loud(
                      'WARNING: Forcing write support on a platform with '
                      'broken mmap semantics in conjunction with direct_io, a '
                      'required FUSE feature. Applications using mmap will '
                      'not function properly, and file corruption may '
                      'result.'
                    )
                elif PLATFORM not in MMAP_KNOWN_GOOD_PLATFORMS:
                    log_loud(
                      'WARNING: Forcing write support on a platform with '
                      'unknown mmap semantics in conjunction with direct_io, '
                      'a required FUSE feature. If mmap implementation is '
                      'broken (and it is on some platforms), file corruption '
                      'may result. Please report success or failure.'
                    )
                else:
                    log_warning(
                      'force_write_support on platform with known-good mmap '
                      'implementation does nothing'
                    )
            else:
                if PLATFORM not in MMAP_KNOWN_GOOD_PLATFORMS:
                    if PLATFORM in MMAP_KNOWN_BAD_PLATFORMS:
                        raise OptParseError(
                          ('This platform is known to have broken mmap '
                           'semantics in conjunction with direct_io, a '
                           'required FUSE feature. You must mount read-only '
                           'with -o %s or use -o force_write_support, but '
                           'forcing write support is *not* recommended.'
                          ) % READ_ONLY_MOUNT_OPTION
                        )
                    else:
                        log_loud(
                          ('WARNING: This platform is not known to have a '
                           'semantically correct mmap implementation in '
                           'conjunction with direct_io, a required FUSE '
                           'feature. You must mount read-only with -o %s or '
                           'use -o force_write_support, but forcing write '
                           'support should only be done judiciously.'
                          ) % READ_ONLY_MOUNT_OPTION
                        )

        return source_tree_rep_cls(
          meta_store = meta_store,
          substitution_patterns = substitution_patterns,
          path_store = path_store_cls(),
          source_tree = SourceTree(
            root = self.source_tree_path,
            iocharset = self.source_iocharset,
          ),
          monitor = monitor,
          cache = cache,
          filters = self.path_filters,
          debug = self.debug,
        )
