# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os

from pytagsfs.util import unicode_path_sep
from pytagsfs.debug import log_debug, log_warning, log_traceback


class SourceTree(object):
    root = None
    iocharset = None

    def __init__(self, root, iocharset = 'utf-8'):
        if not os.path.isabs(root):
            raise ValueError('root "%s" is not an absolute path' % root)
        self.root = root
        self.iocharset = iocharset

    def _validate_path(self, path):
        if not path.startswith(unicode_path_sep):
            raise ValueError(u'%s must start with a %s' % (
              repr(path), unicode_path_sep))

        if (path != unicode_path_sep) and path.endswith(unicode_path_sep):
            raise ValueError(u'%s must not end with a %s' % (
              repr(path), unicode_path_sep))

    def get_relative_path(self, abs_path):
        self._validate_path(abs_path)
        if self.root == unicode_path_sep:
            return abs_path
        rel_path = abs_path[len(self.root):]
        if not rel_path:
            rel_path = unicode_path_sep
        return rel_path

    def get_absolute_path(self, rel_path):
        self._validate_path(rel_path)

        if rel_path == unicode_path_sep:
            return self.root

        abs_path = os.path.join(
          self.root,
          rel_path.lstrip(unicode_path_sep)
        )
        return abs_path

    def walk(self, path = None):
        if path is None:
            path = self.root

        # os.walk seems to want encoded input
        path = self.encode(path)

        for dirpath, dirnames, filenames in os.walk(path):
            dirpath = self.decode(dirpath)
            dirnames = [self.decode(dirname) for dirname in dirnames]
            filenames = [self.decode(filename) for filename in filenames]
            yield (dirpath, dirnames, filenames)

    def decode(self, path):
        try:
            return path.decode(self.iocharset)
        except UnicodeDecodeError:
            log_traceback()
        except UnicodeEncodeError:
            log_traceback()
            if type(path) == unicode:
                log_warning('Warning: tried to decode a unicode string')
        decoded_path = path.decode(self.iocharset, 'replace')
        log_debug(u'Decoded path is "%s".' % decoded_path)
        return decoded_path

    def encode(self, path):
        try:
            return path.encode(self.iocharset)
        except UnicodeEncodeError:
            log_traceback()
        except UnicodeDecodeError:
            log_traceback()
            if type(path) == str:
                log_warning('Warning: tried to encode a byte string')
        encoded_path = path.encode(self.iocharset, 'replace')
        log_debug(u'Encoded path is "%s".' % encoded_path)
        return encoded_path

    def isreadable(self, path):
        try:
            f = open(path, 'r')
            f.read(1)
            f.close()
        except (IOError, OSError):
            return False
        return True
