# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import os, errno, fuse, stat, traceback, sys
from threading import RLock

try:
    from functools import wraps
except ImportError:
    from sclapp.legacy_support import wraps

from pytagsfs.util import (
  now,
  nonatomic,
  attr_named,
  stat_result_to_fuse_stat,
)
from pytagsfs.exceptions import PathNotFound
from pytagsfs.specialfile import SpecialFile

def invalid_argument(*args, **kwargs):
    return -errno.EINVAL

def noop(*args, **kwargs):
    pass

def try_log_file_if_not_found(wrapped):
    @wraps(wrapped)
    def fn(*args, **kwargs):
        try:
            retval = wrapped(*args, **kwargs)
        except PathNotFound:
            retval = -errno.ENOENT
        if retval != -errno.ENOENT:
            return retval
        return getattr(VirtualLogFile, wrapped.__name__)(*args, **kwargs)
    return fn

def check_path(wrapped):
    @wraps(wrapped)
    def fn(self, path, *args, **kwargs):
        if path != os.path.join(os.path.sep, self.filename):
            return -errno.ENOENT
        return wrapped(self, path, *args, **kwargs)
    return fn

class RingCharacterBuffer(object):
    max_length = None
    buffer = None
    len = None

    def __init__(self, max_length):
        self.max_length = max_length
        self.buffer = []
        self.len = 0

    def write(self, s):
        if not isinstance(s, str):
            raise AssertionError(
              'RingCharacterBuffer.write: argument should be a byte string.')

        self.buffer.extend(list(s))
        self.len = self.len + len(s)

        if self.len > self.max_length:
            del self.buffer[0:self.len - self.max_length]
            self.len = self.max_length
            self.__class__ = FullRingCharacterBuffer

    def tell(self):
        return self.len

    def getvalue(self):
        return ''.join(self.buffer)

class FullRingCharacterBuffer(RingCharacterBuffer):
    def write(self, s):
        if not isinstance(s, str):
            raise AssertionError(
              'RingCharacterBuffer.write: argument should be a byte string.')

        self.buffer.extend(list(s))
        del self.buffer[0:len(s)]

class VirtualLogFile(SpecialFile):
    # class
    lock = RLock()

    filename = '.log'
    encoding = 'utf-8'
    file_obj = RingCharacterBuffer(1024 * 1024)

    init_time = now()
    atime = init_time
    mtime = init_time
    ctime = init_time

    # instance
    direct_io = False
    keep = False
    keep_cache = False

################################################################################

    @classmethod
    def set_max_length(cls, bytes):
        cls.file_obj.max_length = bytes

    @classmethod
    @nonatomic(attr_named('lock'))
    def log_write(cls, s):
        if isinstance(s, unicode):
            s = s.encode(cls.encoding)
        cls.file_obj.write(s)
        cls.mtime = now()

    @classmethod
    @check_path
    @nonatomic(attr_named('lock'))
    def getattr(cls, path):
        root_statinfo = cls.filesystem.getattr(os.path.sep)
        try:
            st_dev = root_statinfo.st_dev
            st_uid = root_statinfo.st_uid
            st_gid = root_statinfo.st_gid
        except AttributeError:
            # root_statinfo is actually an integer indicating an error
            from pytagsfs.debug import log_error
            log_error(
              u'VirtualLogFile.getattr: root_statinfo = %s' %
              repr(root_statinfo)
            )
            return root_statinfo

        return stat_result_to_fuse_stat(os.stat_result((
          (stat.S_IFREG | 0444),    #st_mode
          0,                        #st_ino
          st_dev,                   #st_dev
          1,                        #st_link
          st_uid,                   #st_uid
          st_gid,                   #st_gid
          cls.file_obj.tell(),      #st_size
          cls.atime,                #st_atime
          cls.mtime,                #st_mtime
          cls.ctime,                #st_ctime
        )))

    chmod = classmethod(check_path(invalid_argument))
    chmod = classmethod(check_path(invalid_argument))
    chown = classmethod(check_path(invalid_argument))
    truncate = classmethod(check_path(invalid_argument))

    @classmethod
    @check_path
    def open(cls, path, flags):
        return cls(path, flags)

    create = classmethod(check_path(invalid_argument))

    @classmethod
    @check_path
    def utime(cls, path, times):
        try:
            cls.atime = times[0]
        except (TypeError, ValueError):
            cls.atime = now()

    statfs = classmethod(invalid_argument)

    @classmethod
    def readdir(cls, path, offset):
        if path == os.path.sep:
            return [fuse.Direntry(cls.filename)]
        return -errno.ENOENT

    @classmethod
    def access(cls, path, mode):
        cls.atime = now()

    # Undefined FUSE methods:
    #
    # * setxattr
    # * getxattr
    # * listxattr
    # * removexattr
    # * opendir
    # * releasedir
    # * fsyncdir
    # * init
    # * destroy
    # * utimens
    # * bmap
    # * readlink
    # * mknod
    # * mkdir
    # * unlink
    # * symlink
    # * rename
    # * link

################################################################################

    # Undefined File methods:
    #
    # * lock

    def __init__(self, path, flags, mode = None):
        self.path = path
        self.flags = flags
        self.mode = mode

    def fgetattr(self):
        try:
            return self.__class__.getattr(self.filename)
        except:
            print >>sys.stderr, traceback.format_exc()
            raise

    # self.file_obj.getvalue is nonatomic
    @nonatomic(attr_named('lock'))
    def read(self, length, offset):
        try:
            return self.file_obj.getvalue()[offset:offset+length]
        except:
            print >>sys.stderr, traceback.format_exc()
            raise

    write = classmethod(invalid_argument)
    ftruncate = classmethod(invalid_argument)
    fsync = classmethod(noop)
    flush = classmethod(noop)
    release = classmethod(noop)
