# coding: utf-8

# Copyright (c) 2007-2008 Forest Bond.
# This file is part of the pytagsfs software package.
#
# pytagsfs is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# A copy of the license has been included in the COPYING file.

import sys, os, time, traceback, stat, shutil, errno

from sclapp.util import importName
from sclapp.shell import Shell, CommandFailed

from pytagsfs.fs import (
  MMAP_KNOWN_BAD_PLATFORMS,
  READ_ONLY_MOUNT_OPTION,
)
from pytagsfs.metastore.mutagen_ import MutagenFileMetaStore

from common import (
  PLATFORM,
  NO_INOTIFY_PLATFORMS,
  NO_GAMIN_PLATFORMS,
  NO_KQUEUE_PLATFORMS,
  DATA_DIR,
  TestWithDir,
  _UnicodePathsMixin,
  mixin_unicode,
)
from manager import manager


if PLATFORM in MMAP_KNOWN_BAD_PLATFORMS:
    print >>sys.stderr, (
      'warning: platform is known to have a broken mmap implementation, '
      'skipping tests that require write support.'
    )
    platform_supports_writes = False
else:
    platform_supports_writes = True


RACE_DELAY = 2


class ErrorOnError(Exception):
    def __init__(self, original_traceback, new_traceback):
        self.original_traceback = original_traceback
        self.new_traceback = new_traceback
        Exception.__init__(self, original_traceback, new_traceback)

    def __str__(self):
        return '\n'.join([
          '*** Caught an exception while cleaning up after an exception. ***',
          '',
          '=== Traceback for original exception ===',
          self.original_traceback,
          '',
          '=== Traceback for new exception ===',
          self.new_traceback,
        ])

    mount_parameters = None


def add_blackbox_test_class(cls):
    if (not platform_supports_writes) and cls.writes_to_dest_tree:
        print >>sys.stderr, (
          'warning: skipping %s' % cls.__name__)
        return
    manager.add_test_case_class(cls)


class AudioFormatMixin(object):
    def assertFilesHaveSameAudioContent(self, filename1, filename2):
        if self.get_audio_data(filename1) != self.get_audio_data(filename2):
            raise AssertionError

    def get_audio_data(self, filename):
        f = open(filename, 'rb')
        try:
            return self.decode_audio_data(f.read())
        finally:
            f.close()

    def decode_audio_data(self, data):
        stdin, stdout = os.popen2(self.decode_cmd)
        try:
            try:
                stdin.write(data)
            finally:
                stdin.close()
            return stdout.read()
        finally:
            stdout.close()

    def p(self, s):
        s = super(AudioFormatMixin, self).p(s)
        return s.replace(u'ext', self.ext)


class OggMixin(AudioFormatMixin):
    ext = 'ogg'
    decode_cmd = 'ogg123 --quiet --device raw --file - -'


class FlacMixin(AudioFormatMixin):
    ext = 'flac'
    decode_cmd = 'flac --silent --decode --stdout -'


class Mp3Mixin(AudioFormatMixin):
    ext = 'mp3'
    decode_cmd = 'madplay --quiet --output raw:- -'


class _BaseBlackboxTestCase(TestWithDir):
    test_dir_prefix = 'int'

    sh = None

    python_executable = None
    pytagsfs_executable = None
    pytags_executable = None

    umount_cmd = None

    # Indicates that the TestCase writes to the source tree while mounted:
    writes_to_source_tree = False

    # Indicates that the TestCase writes to the dest tree while mounted:
    writes_to_dest_tree = False

    # Indicates that the TestCase requires that the system has non-broken mmap
    # semantics:
    requires_nonbroken_mmap = False

    test_method_name = None

    def __init__(self, methodName = 'runTest'):
        # The unittest module in Python 2.4 saves methodName as
        # __testMethodName, but in Python 2.5, it is _testMethodName.  We skirt
        # this incompatible change by saving it under our own name,
        # test_method_name.
        self.test_method_name = methodName

        super(_BaseBlackboxTestCase, self).__init__(
          methodName = methodName)

    def setUp(self):
        super(_BaseBlackboxTestCase, self).setUp()

        package_root = os.path.dirname(
          os.path.dirname(os.path.abspath(__file__)))
        modules_root = os.path.join(package_root, 'modules')

        shell = '/bin/sh'

        import commands
        status, output = commands.getstatusoutput('%s --version' % shell)
        if (status == 0) and ('bash' in output):
            shell = '%s --posix --noediting' % shell

        self.sh = Shell(shell = shell, delaybeforesend = 0)

        self.sh.env['PATH'] = '%s:%s' % (package_root, self.sh.env['PATH'])
        self.sh.env['PYTHONPATH'] = '%s:%s' % (
          modules_root, self.sh.env['PYTHONPATH'])

        self.python_executable = sys.executable
        self.pytagsfs_executable = os.path.join(package_root, 'pytagsfs')
        self.pytags_executable = os.path.join(package_root, 'pytags')

        status, output = commands.getstatusoutput(
          'test -x "$(which fusermount)"')
        if status == 0:
            self.umount_cmd = 'fusermount -u'
        else:
            status, output = commands.getstatusoutput(
              'test -x "$(which umount)"')
            if status == 0:
                self.umount_cmd = 'umount'
            else:
                raise AssertionError(
                  'no valid umount command could be determined')

    def tearDown(self):
        self.sh.exit()
        del self.sh
        super(_BaseBlackboxTestCase, self).tearDown()

    def p(self, s):
        return unicode(s)

    def get_audio_data(self, filename):
        raise NotImplementedError

    def get_tags(self, filename):
        return MutagenFileMetaStore.tags_class(filename)

    def get_tag(self, filename, tag):
        return self.get_tags(filename)[tag]

    def set_tag(self, filename, tag, value):
        tags = self.get_tags(filename)
        tags[tag] = value
        tags.save()

    def assertFilesHaveSameTags(self, filename1, filename2):
        self.assertEqual(self.get_tags(filename1), self.get_tags(filename2))

    def get_opts(self):
        raise NotImplementedError

    def get_mount_parameters(self):
        return {}

    def mount(self):
        mount_parameters = dict(self.get_mount_parameters())

        extra_options = []
        if mount_parameters.get('sourcetreerep', None):
            extra_options.append(
              'sourcetreerep=%s' % mount_parameters['sourcetreerep'])
        if mount_parameters.get('sourcetreemon', None):
            extra_options.append(
              'sourcetreemon=%s' % mount_parameters['sourcetreemon'])
        if mount_parameters.get('pathstore', None):
            extra_options.append(
              'pathstore=%s' % mount_parameters['pathstore'])

        extra_options.append(self.get_opts())

        if not platform_supports_writes:
            extra_options.append(READ_ONLY_MOUNT_OPTION)

        # Some systems (Mac OSX, for example), have somewhat small limitations
        # on the length of mount point paths.  This caused problems in the past
        # when the test name (which can be quite long) was used as the
        # directory name.  Consequently, our test directory ($PWD, here,
        # created via TestWithDir) must have a reasonably short path.

        self.sh.execute(
          'mount_cmd="? ? -o debug,? source \'${PWD}/mnt\'"',
          self.python_executable,
          self.pytagsfs_executable,
          ','.join(extra_options),
        )
        self.sh.execute('echo "${mount_cmd}" >mount_cmd')
        self.sh.execute(
          'eval $mount_cmd >logfile 2>&1 &',
        )

        time.sleep(RACE_DELAY)

        # Check that mnt is accessible:
        os.stat('mnt')

    def umount(self):
        try:
            self.sh.execute('%s "${PWD}/mnt"' % self.umount_cmd)
        except CommandFailed:
            time.sleep(RACE_DELAY)
            self.sh.execute('%s "${PWD}/mnt"' % self.umount_cmd)

    def _init(self):
        mount_parameters = self.get_mount_parameters()

        test_name = '%s.%s' % (self.__class__.__name__, self.test_method_name)
        self.test_name_file = os.path.join(self.test_dir, 'test_name')

        # Store the test name in a file so that the test tree can be linked
        # to the test that failed for debugging.
        self.sh.execute('echo ? >?', test_name, self.test_name_file)

        self.wd = os.getcwd()
        os.chdir(self.test_dir)
        self.sh.pushd(self.test_dir)

        self.sh.execute('mkdir -p mnt source')
        self.build_tree()

    def _de_init(self):
        # Check that mnt is accessible:
        os.stat('mnt')

        self.umount()
        self.sh.execute('rm -f mount_cmd')
        self.sh.execute('rm -f logfile')
        self.sh.execute('rm -R mnt source')
        self.sh.popd()
        self.sh.execute('rm ?', self.test_name_file)
        os.chdir(self.wd)

    def _de_init_exc(self):
        try:
            try:
                original_traceback = traceback.format_exc()
                try:
                    self.record_failure()
                except Exception:
                    print >>sys.stderr, traceback.format_exc()
                try:
                    self.umount()
                except Exception:
                    new_traceback = traceback.format_exc()
                    raise ErrorOnError(
                      original_traceback,
                      new_traceback,
                    )
                raise
            finally:
                self.sh.popd()
        finally:
            os.chdir(self.wd)

    def record_failure(self):
        try:
            self.sh.execute(
              '( find . -print0 | xargs -0 -n1 ls -lhd ) >listing 2>&1')
        except CommandFailed:
            print >>sys.stderr, traceback.format_exc()

    def assertFileExists(self, filename):
        try:
            f = open(filename, 'r')
        except (OSError, IOError), e:
            raise AssertionError(str(e))
        else:
            f.close()

    def assertFileDoesNotExist(self, filename):
        def test_fn():
            f = open(filename, 'r')
            f.close()
        self.assertRaises(IOError, test_fn)

    def assertFileContents(self, filename, expected):
        f = open(filename, 'r')
        try:
            actual = f.read().decode('utf-8', 'replace')
        finally:
            f.close()
        assert (actual == expected), (
          u'Expected:\n%s\nGot:\n%s' % (expected, actual)
        )

    def assertFileContains(self, filename, s):
        f = open(filename, 'r')
        try:
            contents = f.read().decode('utf-8', 'replace')
        finally:
            f.close()
        assert (s in contents), (
          u'Expected to find:\n%s\nGot:\n%s' % (s, contents)
        )

    def _assertFilePredicate(self, filename1, filename2, predicate):
        f1 = open(filename1, 'r')
        try:
            f2 = open(filename2, 'r')
            try:
                assert predicate(f1, f2)
            finally:
                f2.close()
        finally:
            f1.close()

    def assertFilesAreSame(self, filename1, filename2):
        self._assertFilePredicate(
          filename1, filename2, (lambda f1, f2: f1.read() == f2.read()))

    def assertFilesAreDifferent(self, filename1, filename2):
        self._assertFilePredicate(
          filename1, filename2, (lambda f1, f2: f1.read() != f2.read()))

    def assertDirectoryExists(self, dirname):
        try:
            os.listdir(dirname)
        except (IOError, OSError), e:
            raise AssertionError(unicode(e))

    def assertDirectoryDoesNotExist(self, dirname):
        def test_fn():
            os.listdir(dirname)
        self.assertRaises((IOError, OSError), test_fn)


################################################################################


class _BasePathPatternBlackboxTestCase(_BaseBlackboxTestCase):
    def build_tree(self):
        self.sh.execute(u'mkdir -p source/a/b source/a source/x')
        self.sh.execute(u'touch source/a/b/c source/a/• source/x/y')
        self.sh.execute(u'chmod 0644 source/a/b/c source/a/• source/x/y')


class SimpleFilenamePatternTestCase(_BasePathPatternBlackboxTestCase):
    writes_to_source_tree = True

    def get_opts(self):
        return 'format=/%f'

    def test_read_dir(self):
        self._init()
        self.mount()
        try:
            self.assertEqual(
              set(f for f in os.listdir(u'mnt') if not f.startswith('.')),
              set([u'c', u'y', u'•']),
            )
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init()

    def test_add_empty_file(self):
        self._init()
        self.mount()
        try:
            self.sh.execute('touch source/a/b/d')
            time.sleep(RACE_DELAY)
            self.assertFileExists('mnt/d')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init()

add_blackbox_test_class(SimpleFilenamePatternTestCase)


class SourceExcludeTestCase(_BasePathPatternBlackboxTestCase):
    def get_opts(self):
        return 'format=/%f,srcfilter=!\/c$'

    def test_source_exclusion(self):
        self._init()
        self.mount()
        try:
            filenames = os.listdir(u'mnt')
            assert (u'c' not in filenames), repr(filenames)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(SourceExcludeTestCase)


class DestExcludeTestCase(_BasePathPatternBlackboxTestCase):
    def get_opts(self):
        return 'format=/%f,srcfilter=!\/c$'

    def test_dest_exclusion(self):
        self._init()
        self.mount()
        try:
            filenames = os.listdir(u'mnt')
            assert (u'c' not in filenames), repr(filenames)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(DestExcludeTestCase)


class _BasePathPatternUpdatesTestCase(_BasePathPatternBlackboxTestCase):
    def get_opts(self):
        return 'format=/%p/%f'


class PathPatternSourceTreeUpdatesTestCase(_BasePathPatternUpdatesTestCase):
    writes_to_source_tree = True

    def test_update_file_contents(self):
        self._init()
        self.mount()
        try:
            content = 'foo\n'

            f = open('source/a/b/c', 'w')
            f.write(content)
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', content)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_update_mtime(self):
        self._init()
        self.mount()
        try:
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            stat_a_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_a_src.st_mtime, stat_a_dst.st_mtime)
            self.assertEqual(stat_a_src.st_mtime, stat_a_dst_dir.st_mtime)

            # Make sure enough time passes for timestamps to change.
            time.sleep(1)

            os.utime('source/a/b/c', None)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            stat_b_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_b_src.st_mtime, stat_b_dst.st_mtime)
            self.assertEqual(stat_b_src.st_mtime, stat_b_dst_dir.st_mtime)

            assert stat_a_src.st_mtime < stat_b_src.st_mtime
            assert stat_a_dst.st_mtime < stat_b_dst.st_mtime
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_update_mode(self):
        self._init()
        self.mount()
        try:
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            mode_a_src = stat.S_IMODE(stat_a_src.st_mode)
            mode_a_dst = stat.S_IMODE(stat_a_dst.st_mode)

            self.assertEqual(mode_a_src, 0644)
            self.assertEqual(mode_a_dst, 0644)

            new_mode = (mode_a_src & 0000) | 0600

            os.chmod('source/a/b/c', new_mode)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            mode_b_src = stat.S_IMODE(stat_b_src.st_mode)
            mode_b_dst = stat.S_IMODE(stat_b_dst.st_mode)

            self.assertEqual(mode_b_src, 0600)
            self.assertEqual(mode_b_dst, 0600)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_simple_rename(self):
        self._init()
        self.mount()
        try:
            os.makedirs('source/m/n')
            os.rename('source/a/b/c', 'source/m/n/o')

            time.sleep(RACE_DELAY)

            self.assertFileDoesNotExist('mnt/b/c')
            self.assertFileExists('mnt/n/o')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(PathPatternSourceTreeUpdatesTestCase)


class PathPatternDestTreeUpdatesTestCase(_BasePathPatternUpdatesTestCase):
    writes_to_dest_tree = True

    def test_write_truncate(self):
        self._init()
        self.mount()
        try:
            f = open('mnt/b/c', 'w')
            f.write('foo\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\n')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_append(self):
        self._init()
        self.mount()
        try:
            f = open('mnt/b/c', 'w')
            f.write('foo\nbar\nbaz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\n')

            f = open('mnt/b/c', 'a')
            f.write('buz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\nbuz\n')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_overwrite_section(self):
        self._init()
        self.mount()
        try:
            f = open('mnt/b/c', 'w')
            f.write('foo\nbar\nbaz\n')
            f.close()

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbar\nbaz\n')

            fd = os.open('mnt/b/c', os.O_RDWR)
            os.lseek(fd, 4, 0)
            os.write(fd, 'biz\n')
            os.close(fd)

            time.sleep(RACE_DELAY)

            self.assertFileContents('mnt/b/c', 'foo\nbiz\nbaz\n')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_ftruncate(self):
        initial = 'foo\nbar\nbaz\nbiz\nbang\nboom'
        final = initial[:8]

        self._init()

        f = open('source/a/m', 'w')
        f.write(initial)
        f.close()
        time.sleep(RACE_DELAY)

        self.mount()
        try:
            self.assertFileContents('mnt/a/m', initial)

            f = open('mnt/a/m', 'r+')
            try:
                self.assertEqual(f.read(), initial)
                f.truncate(8)
                f.seek(0)
                self.assertEqual(f.read(), final)
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', final)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_seek_to_end(self):
        content = 'foo\nbar\nbaz\nbiz\nbang\nboom'

        self._init()

        f = open('source/a/m', 'w')
        try:
            f.write(content)
        finally:
            f.close()
        time.sleep(RACE_DELAY)

        self.mount()
        try:
            self.assertFileContents('mnt/a/m', content)

            f = open('mnt/a/m', 'rb+')
            try:
                self.assertEqual(f.read(), content)
                f.seek(0, 2)
                self.assertEqual(f.tell(), len(content))
            finally:
                f.close()

            self.assertFileContents('mnt/a/m', content)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_write_past_end_and_seek_to_end(self):
        content = 'foo\nbar\n'
        last = 'baz\n'
        final = content + last

        self._init()

        f = open('source/a/m', 'w')
        try:
            f.write(content)
        finally:
            f.close()
        time.sleep(RACE_DELAY)

        self.mount()
        try:
            self.assertFileContents('mnt/a/m', content)

            f = open('mnt/a/m', 'rb+')
            try:
                self.assertEqual(f.read(), content)
                f.seek(0, 2)
                f.write(last)
                f.seek(0)
                self.assertEqual(f.read(), final)
                f.seek(0, 2)
                self.assertEqual(f.tell(), len(final))
            finally:
                f.close()

            time.sleep(RACE_DELAY)
            self.assertFileContents('mnt/a/m', final)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_update_mtime(self):
        self._init()
        self.mount()
        try:
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            stat_a_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_a_src.st_mtime, stat_a_dst.st_mtime)
            self.assertEqual(stat_a_src.st_mtime, stat_a_dst_dir.st_mtime)

            # Make sure enough time passes for timestamps to change.
            time.sleep(1)

            os.utime('mnt/b/c', None)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            stat_b_dst_dir = os.stat('mnt/b')

            self.assertEqual(stat_b_src.st_mtime, stat_b_dst.st_mtime)
            self.assertEqual(stat_b_src.st_mtime, stat_b_dst_dir.st_mtime)

            assert stat_a_src.st_mtime < stat_b_src.st_mtime
            assert stat_a_dst.st_mtime < stat_b_dst.st_mtime
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_update_mode(self):
        self._init()
        self.mount()
        try:
            stat_a_src = os.stat('source/a/b/c')
            stat_a_dst = os.stat('mnt/b/c')

            mode_a_src = stat.S_IMODE(stat_a_src.st_mode)
            mode_a_dst = stat.S_IMODE(stat_a_dst.st_mode)

            self.assertEqual(mode_a_src, 0644)
            self.assertEqual(mode_a_dst, 0644)

            new_mode = (mode_a_src & 0000) | 0600

            os.chmod('mnt/b/c', new_mode)

            time.sleep(RACE_DELAY)

            stat_b_src = os.stat('source/a/b/c')
            stat_b_dst = os.stat('mnt/b/c')

            mode_b_src = stat.S_IMODE(stat_b_src.st_mode)
            mode_b_dst = stat.S_IMODE(stat_b_dst.st_mode)

            self.assertEqual(mode_b_src, 0600)
            self.assertEqual(mode_b_dst, 0600)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(PathPatternDestTreeUpdatesTestCase)


################################################################################


class _BaseSourceTreeUpdatesTestCase(_BaseBlackboxTestCase):
    writes_to_source_tree = True

    def get_opts(self):
        return 'format=/%{artist} - %t.%e'

    def build_tree(self):
        self.data_file = self.p(os.path.join(DATA_DIR, 'silence.ext'))
        shutil.copy(self.data_file, self.p('source/foo.ext'))
        self.set_tag(self.p('source/foo.ext'), 'artist', self.p('bar'))
        self.set_tag(self.p('source/foo.ext'), 'title', self.p('baz'))

    def test_tag_change_causing_path_change(self):
        self._init()
        self.mount()
        try:
            self.set_tag(self.p('source/foo.ext'), 'artist', self.p('qux'))
            time.sleep(RACE_DELAY)
            self.assertEqual(
              self.get_tag(self.p('mnt/qux - baz.ext'), 'artist'),
              [self.p('qux')],
            )
            self.assertFilesHaveSameAudioContent(
              self.data_file, self.p('mnt/qux - baz.ext'))
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_tag_change_causing_new_fake_path(self):
        self._init()
        self.set_tag(self.p('source/foo.ext'), 'artist', [])
        self.mount()
        try:
            self.assertEqual(os.listdir('mnt'), ['.log'])
            self.set_tag(self.p('source/foo.ext'), 'artist', self.p('qux'))
            time.sleep(RACE_DELAY)
            self.assertEqual(
              self.get_tag(self.p('mnt/qux - baz.ext'), 'artist'),
              [self.p('qux')],
            )
            self.assertFilesHaveSameAudioContent(
              self.data_file, self.p('mnt/qux - baz.ext'))
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 


class OggSourceTreeUpdatesTestCase(OggMixin, _BaseSourceTreeUpdatesTestCase):
    pass

add_blackbox_test_class(OggSourceTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(OggSourceTreeUpdatesTestCase))


class FlacSourceTreeUpdatesTestCase(FlacMixin, _BaseSourceTreeUpdatesTestCase):
    pass

add_blackbox_test_class(FlacSourceTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(FlacSourceTreeUpdatesTestCase))


class Mp3SourceTreeUpdatesTestCase(Mp3Mixin, _BaseSourceTreeUpdatesTestCase):
    pass

add_blackbox_test_class(Mp3SourceTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(Mp3SourceTreeUpdatesTestCase))


class _BaseDestTreeUpdatesTestCase(_BaseBlackboxTestCase):
    writes_to_dest_tree = True

    def get_opts(self):
        return 'format=/%{artist} - %t.%e'

    def build_tree(self):
        self.data_file = self.p(os.path.join(DATA_DIR, 'silence.ext'))
        shutil.copy(self.data_file, self.p('source/foo.ext'))
        self.set_tag(self.p('source/foo.ext'), 'artist', self.p('bar'))
        self.set_tag(self.p('source/foo.ext'), 'title', self.p('baz'))

    def test_tag_change_causing_path_change(self):
        self._init()
        self.mount()
        try:
            self.set_tag(self.p('mnt/bar - baz.ext'), 'artist', self.p('qux'))
            time.sleep(RACE_DELAY)
            self.assertEqual(
              self.get_tag(self.p('source/foo.ext'), 'artist'),
              [self.p('qux')],
            )
            self.assertFilesHaveSameAudioContent(
              self.data_file, self.p('mnt/qux - baz.ext'))
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_remove_file(self):
        self._init()
        self.mount()
        try:
            os.unlink(self.p('source/foo.ext'))
            time.sleep(RACE_DELAY)
            self.assertEqual(os.listdir('mnt'), ['.log'])
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 


class OggDestTreeUpdatesTestCase(OggMixin, _BaseDestTreeUpdatesTestCase):
    pass

add_blackbox_test_class(OggDestTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(OggDestTreeUpdatesTestCase))


class FlacDestTreeUpdatesTestCase(FlacMixin, _BaseDestTreeUpdatesTestCase):
    pass

add_blackbox_test_class(FlacDestTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(FlacDestTreeUpdatesTestCase))


class Mp3DestTreeUpdatesTestCase(Mp3Mixin, _BaseDestTreeUpdatesTestCase):
    pass

add_blackbox_test_class(Mp3DestTreeUpdatesTestCase)
add_blackbox_test_class(mixin_unicode(Mp3DestTreeUpdatesTestCase))


################################################################################


class _BaseRenameTestCase(_BaseBlackboxTestCase):
    writes_to_dest_tree = True

    def build_tree(self):
        shutil.copy(
          self.p(os.path.join(DATA_DIR, 'silence.ext')),
          self.p('source/foo.ext'),
        )
        self.set_tag(self.p('source/foo.ext'), 'artist', self.p('bar'))
        self.set_tag(self.p('source/foo.ext'), 'title', self.p('baz'))

    def get_opts(self):
        return 'format=/%a/%t.%e'

    def test_simple_rename(self):
        self._init()
        self.mount()
        try:
            os.rename(
              self.p(os.path.join('mnt', 'bar', 'baz.ext')),
              self.p(os.path.join('mnt', 'bar', 'qux.ext')),
            )
            time.sleep(RACE_DELAY)
            self.assertFilesHaveSameAudioContent(
              self.p(os.path.join(DATA_DIR, 'silence.ext')),
              self.p(os.path.join('mnt', 'bar', 'qux.ext')),
            )
            tags = self.get_tags(self.p(os.path.join('mnt', 'bar', 'qux.ext')))
            self.assertEqual(tags['artist'], [self.p('bar')])
            self.assertEqual(tags['title'], [self.p('qux')])
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_rename_across_directories(self):
        self._init()
        self.mount()
        try:
            os.mkdir(self.p(os.path.join('mnt', 'qux')))
            os.rename(
              self.p(os.path.join('mnt', 'bar', 'baz.ext')),
              self.p(os.path.join('mnt', 'qux', 'baz.ext')),
            )
            time.sleep(RACE_DELAY)
            self.assertFilesHaveSameAudioContent(
              self.p(os.path.join(DATA_DIR, 'silence.ext')),
              self.p(os.path.join('mnt', 'qux', 'baz.ext')),
            )
            tags = self.get_tags(self.p(os.path.join('mnt', 'qux', 'baz.ext')))
            self.assertEqual(tags['artist'], [self.p('qux')])
            self.assertEqual(tags['title'], [self.p('baz')])
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_directory_rename(self):
        self._init()
        self.mount()
        try:
            os.rename(
              self.p(os.path.join('mnt', 'bar')),
              self.p(os.path.join('mnt', 'qux')),
            )
            time.sleep(RACE_DELAY)
            self.assertFilesHaveSameAudioContent(
              self.p(os.path.join(DATA_DIR, 'silence.ext')),
              self.p(os.path.join('mnt', 'qux', 'baz.ext')),
            )
            tags = self.get_tags(self.p(os.path.join('mnt', 'qux', 'baz.ext')))
            self.assertEqual(tags['artist'], [self.p('qux')])
            self.assertEqual(tags['title'], [self.p('baz')])
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_invalid_rename(self):
        self._init()
        self.mount()
        try:
            try:
                os.rename(
                  self.p(os.path.join('mnt', 'bar', 'baz.ext')),
                  self.p(os.path.join('mnt', 'bar', 'qux.boink')),
                )
            except OSError, e:
                self.assertEqual(e.errno, errno.EINVAL)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 


class OggRenameTestCase(OggMixin, _BaseRenameTestCase):
    pass

add_blackbox_test_class(OggRenameTestCase)
add_blackbox_test_class(mixin_unicode(OggRenameTestCase))


class FlacRenameTestCase(FlacMixin, _BaseRenameTestCase):
    pass

add_blackbox_test_class(FlacRenameTestCase)
add_blackbox_test_class(mixin_unicode(FlacRenameTestCase))


class Mp3RenameTestCase(Mp3Mixin, _BaseRenameTestCase):
    pass

add_blackbox_test_class(Mp3RenameTestCase)
add_blackbox_test_class(mixin_unicode(Mp3RenameTestCase))


################################################################################


class MkdirRmdirTagsTestCase(_BaseBlackboxTestCase):
    writes_to_dest_tree = True

    def get_opts(self):
        return 'format=/%g/%{artist}/%t.%e'

    def build_tree(self):
        data_file = os.path.join(DATA_DIR, 'silence.flac')

        for num in (1, 2, 3, 4):
            self.sh.execute('cp ? ?', data_file, 'source/silence%u.flac' % num)

        for num, artist, title, genre in (
          (1, 'foo', 'bar', 'Rock'),
          (2, 'foo', 'baz', 'Rock'),
          (3, 'bink', 'bonk', 'Rap'),
          (4, 'fink', 'fonk', 'Country'),
        ):
            file = 'source/silence%u.flac' % num
            self.set_tag(file, 'artist', artist)
            self.set_tag(file, 'title', title)
            self.set_tag(file, 'genre', genre)

    def test_mkdir_rmdir(self):
        self._init()
        self.mount()
        try:
            dir = 'mnt/Honky Tonk'
            self.sh.execute('mkdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists(dir)
            self.sh.execute('rmdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_rename_empty_dir(self):
        self._init()
        self.mount()
        try:
            dir = 'mnt/Honky Tonk'
            dir2 = 'mnt/Soul'
            self.sh.execute('mkdir ?', dir)
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists(dir)
            self.sh.execute('mv ? ?', dir, dir2)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir)
            self.sh.execute('ls ?', dir2)
            self.assertRaises(CommandFailed, self.sh.execute, 'ls ?/*', dir2)
            self.sh.execute('rmdir ?', dir2)
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist(dir2)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_move_files_into_new_directory(self):
        self._init()
        self.mount()
        try:
            self.sh.execute('mkdir mnt/Rock/biz')
            time.sleep(RACE_DELAY)
            self.assertDirectoryExists('mnt/Rock/biz')

            self.sh.execute('mv mnt/Rock/foo/* mnt/Rock/biz/')
            time.sleep(RACE_DELAY)
            self.assertDirectoryDoesNotExist('mnt/Rock/foo')

            filenames = os.listdir(u'mnt/Rock/biz')
            self.assertEqual(len(filenames), 2)
            for filename in filenames:
                tags = self.get_tags(os.path.join(u'mnt/Rock/biz', filename))
                self.assertEqual(tags['artist'], [u'biz'])
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(MkdirRmdirTagsTestCase)


################################################################################


class LogFileTestCase(_BasePathPatternBlackboxTestCase):
    def get_opts(self):
        return 'verbosity=debug'

    def test_log_file_exists(self):
        self._init()
        self.mount()
        try:
            self.assertFileExists('mnt/.log')
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

    def test_log_file_receives_messages(self):
        self._init()
        self.mount()
        try:
            from pytagsfs import __version__ as version
            self.assertFileContains('mnt/.log', 'pytagsfs version %s' % version)
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(LogFileTestCase)


class LogSizeTestCase(_BasePathPatternBlackboxTestCase):
    logsize = 8

    def get_opts(self):
        return 'verbosity=debug,logsize=%u' % self.logsize

    def test_log_size(self):
        self._init()
        self.mount()
        try:
            f = open('mnt/.log', 'r')
            length = len(f.read())
            try:
                assert length == self.logsize, (
                  'Expected length to be %u, got %u instead.' % (
                    self.logsize, length))
            finally:
                f.close()
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(LogSizeTestCase)


################################################################################


class MmapTestCase(_BaseBlackboxTestCase):
    writes_to_dest_tree = True

    def get_opts(self):
        return 'format=/%f'

    def build_tree(self):
        f = open('source/foo.txt', 'w')
        f.write('foo\nbar\n')
        f.close()

    def test_mmap_create_fails(self):
        self._init()
        self.mount()
        try:
            try:
                import mmap
            except ImportError:
                return

            f = open('mnt/foo.txt', 'a+')

            try:
                f.seek(0, 2)
                filesize = f.tell()

                try:
                    map = mmap.mmap(f.fileno(), filesize)
                except (EnvironmentError, ValueError):
                    pass
                else:
                    map.close()
                    self.fail('mmap succesfully created (should fail)')
            finally:
                f.close()
        except:
            self._de_init_exc()
            raise
        else:
            self._de_init() 

add_blackbox_test_class(MmapTestCase)
